# This script does two things - it extracts and sets up the daily snow data from the raw SNODAS product, and it creates monthly summary statistic rasters of average depth, std of depth, days with snow >8inches, and number of days with no snow at all.
# The summary monthly rasters are stored in each month/year folder \<YYYY>\<MM_MONTHABR> (e.g. \2009\01_Jan)
# Years in this case are actual year, not water year
# OUTPUTS:
# Raster for each day of VT Snow depth vt_depth_<YYYYMM>.tif (M is always 2 digits)
# Mean depth over all days of month depth_ave_<YYYYM>.tif (M is not always a 2-digit number, Jan is 1, Dec is 12)
# Standard deviation of depth over month  depth_std_<YYYYM>.tif (M is not always a 2-digit number, Jan is 1, Dec is 12)
# Sum of days over month where depth >8 inches mds_<YYYYM>.tif (M is not always a 2-digit number, Jan is 1, Dec is 12)
# Sum of days over month where there was no snow mdns_<YYYYM>.tif (M is not always a 2-digit number, Jan is 1, Dec is 12)
# The monthly summaries are later used by summarySamplingByPoints.py
# Author: Jim Duncan, UVM/FEMC james.duncan@uvm.edu
# Last updated: 9/9/2020

#Forest Water Attribs to capture:
#Maximum snowdepth per snowyear
#Trend in maximum snowdepth over time
#Number of days with snow depth >1 in per year
#Trend in the number of days with snowdepth >1 in (>2.54 cm) over time 
#Duration of snowpack per year
#Trend in snowpack duration
#Average snow water equivilent per snowyear
#Trend in average snow water equivalent


import arcpy
from arcpy import env
from arcpy.sa import *
import os
import csv
import tarfile, gzip

arcpy.CheckOutExtension("Spatial")
#where the snow data is stored
directory="D:/Duncan_SNODAS/snodasData/"
#where to do the processing
workingDirectory="D:/Duncan_SNODAS/snodasProcessing/"
#the shape used to extract the final snow raster for each day
regionClipShape="D:/Duncan_SNODAS/RegionBoundary_projected.shp"
#The years of snodas data to process
years=[2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018,2019,2020]
#The months of SNODAS data to process
months=[1,2,3,4,5,10,11,12]
#the envelope containing vermont, with some wiggle room. This is used to do an initial extraction of the continental dataset so we're only working with a subset to start.
roughRectangle="-84 39 -66 48"
#crosswalks month values to naming conventions in the filesystem
monthDict={1:"01_Jan",2:"02_Feb",3:"03_Mar",4:"04_Apr",5:"05_May",6:"06_Jun",7:"07_Jul",8:"08_Aug",9:"09_Sep",10:"10_Oct",11:"11_Nov",12:"12_Dec"}


#Loops through directory structures, extracts the snow depth and SWE component of the SNODAS data, then projects and extracts to clip boundary and sets nulls
def snodasExtractClip(directory, workingDirectory, theYear, theMonth):

    #get the month name from the dictionary
    theMonth=monthDict[theMonth]
    #For testing
    #theMonth="09_Sep"
    #theYear="2003"
    
    #cast to strings
    theYear=str(theYear)
    theMonth=str(theMonth)
    print "Looking for files in"+directory+" "+theYear+" year and " + theMonth + " month"
    
    
    #build the source path
    thePath=directory + theYear + "/" + theMonth + "/"
    if (os.path.isdir(thePath)!=True):
        print "No directory found for this year and month"
        return
        
    # get the list of tar files in this directory
    arcpy.env.workspace = thePath
    tarFiles = arcpy.ListFiles('*.tar')
    # print tarFiles
    
    # Loop through files and extract, clip and create header
    for tarFile in tarFiles:
        
        thefile=tarfile.open(thePath+tarFile,"r")
        
        #extract the yearmonthday from filename
        datestamp=tarFile[7:-4]
        print "Processing file for "+ datestamp
        
        #File format: rr_mmmffppppSvvvvTttttaaaaTSyyyymmddhhIP00Z.xxx.gz
        #Find info at http://nsidc.org/data/G02158
        #We want the us_ssmv11036tS__T0001TTNATS2003093005HP001 file for snow depth
        #We want the us_ssmv11034tS__T0001TTNATS2003093005HP001 file for SWE

        #set up working directory for this year
        theWorkingPath=workingDirectory+ theYear + "/" + theMonth + "/" 
        if (os.path.isdir(theWorkingPath)!=True):
            os.makedirs(theWorkingPath)
        #change to that directory and set arc workspace
        os.chdir(theWorkingPath)
        env.workspace=theWorkingPath
        

        
        # only move forward with depth if this file hasn't already been done
        if(os.path.isfile("region_depth_"+datestamp+".tif")!=True) :
            #dat and hdr file templates
            datFileName="us_ssmv11036tS__T0001TTNATS"+datestamp + "05HP001.dat"
            hdrFileName="us_ssmv11036tS__T0001TTNATS"+datestamp + "05HP001.Hdr"
            print "extracting file"
            #extract the gz from the tar
            # For a short while the changed the internal structure of the tar, so have to test for that case 
            try:
                thefile.extract(datFileName+".gz")
            except KeyError:
                thefile.extract("./"+datFileName+".gz")
            
            #read the file contents and write to the .dat file
            with gzip.open(datFileName +".gz", 'rb') as f:
                dataFile=open(datFileName,"wb")
                dataFile.write(f.read())
                dataFile.close()
                f.close()
            
            #Manually create the header to work with Arc
            header = "nrows 3351 \nncols 6935 \nnbands 1 \nnbits 16 \npixeltype signedint \nbyteorder M \nlayout dat \nulxmap -124.729583333331703 \nulymap 52.871249516804028 \nxdim 0.00833333333 \nydim 0.00833333333"
            #Read the data from the data file
            headerFile=open(hdrFileName,"w")
            headerFile.write(header)
            headerFile.close()
            
            #define the intermediate step datasets
            clipped="clip_"+datestamp+".tif"
            proj="proj_"+datestamp+".tif"
            regionclip="region_clip_"+datestamp+".tif"
            depth="region_depth_"+datestamp+".tif"
                    
            #define the projection for the data
            
            #print "defining projection"
            arcpy.DefineProjection_management(in_dataset=datFileName, coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]")
            #print "clipping to rough box"
            arcpy.Clip_management(in_raster=datFileName, rectangle=roughRectangle, out_raster=clipped, in_template_dataset="", nodata_value="", clipping_geometry="NONE", maintain_clipping_extent="NO_MAINTAIN_EXTENT")
            #print "projecting to state plane"
            # Replace a layer/table view name with a path to a dataset (which can be a layer file) or create the layer/table view within the script
            arcpy.ProjectRaster_management(in_raster=clipped, out_raster=proj, out_coor_system="PROJCS['USA_Contiguous_Albers_Equal_Area_Conic_USGS_version',GEOGCS['GCS_North_American_1983',DATUM['D_North_American_1983',SPHEROID['GRS_1980',6378137.0,298.257222101]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]],PROJECTION['Albers'],PARAMETER['False_Easting',0.0],PARAMETER['False_Northing',0.0],PARAMETER['Central_Meridian',-96.0],PARAMETER['Standard_Parallel_1',29.5],PARAMETER['Standard_Parallel_2',45.5],PARAMETER['Latitude_Of_Origin',23.0],UNIT['Meter',1.0]]", resampling_type="NEAREST", cell_size="923.166496502035 923.166496502035", geographic_transform="NAD_1983_To_WGS_1984_1", Registration_Point="", in_coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]", vertical="NO_VERTICAL")
            
            #clip it to the extent of interest
            #print "clipping dataset"
            arcpy.gp.ExtractByMask_sa(proj, regionClipShape, regionclip)
            
            print "setting nulls"
            arcpy.gp.SetNull_sa(regionclip, regionclip, depth, '"Value" <0')
            
            #clean up!
            os.remove(datFileName+".gz")
            arcpy.Delete_management(datFileName)
            arcpy.Delete_management(clipped)
            arcpy.Delete_management(proj)
            arcpy.Delete_management(regionclip)
        else:
            print "Skipping depth file for "+ datestamp +", already done"
            
        # only move forward with SWE if this file hasn't already been done
        if(os.path.isfile("region_swe_"+datestamp+".tif")!=True) :
            #dat and hdr file templates
            datFileName="us_ssmv11034tS__T0001TTNATS"+datestamp + "05HP001.dat"
            hdrFileName="us_ssmv11034tS__T0001TTNATS"+datestamp + "05HP001.Hdr"
            print "extracting file"
            #extract the gz from the tar
            # For a short while the changed the internal structure of the tar, so have to test for that case 
            try:
                thefile.extract(datFileName+".gz")
            except KeyError:
                thefile.extract("./"+datFileName+".gz")
            
            #read the file contents and write to the .dat file
            with gzip.open(datFileName +".gz", 'rb') as f:
                dataFile=open(datFileName,"wb")
                dataFile.write(f.read())
                dataFile.close()
                f.close()
            
            #Manually create the header to work with Arc
            header = "nrows 3351 \nncols 6935 \nnbands 1 \nnbits 16 \npixeltype signedint \nbyteorder M \nlayout dat \nulxmap -124.729583333331703 \nulymap 52.871249516804028 \nxdim 0.00833333333 \nydim 0.00833333333"
            #Read the data from the data file
            headerFile=open(hdrFileName,"w")
            headerFile.write(header)
            headerFile.close()
            
            #define the intermediate step datasets
            clipped="clip_"+datestamp+".tif"
            proj="proj_"+datestamp+".tif"
            regionclip="region_clip_"+datestamp+".tif"
            swe="region_swe_"+datestamp+".tif"
                    
            #define the projection for the data
            
            #print "defining projection"
            arcpy.DefineProjection_management(in_dataset=datFileName, coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]")
            #print "clipping to rough box"
            arcpy.Clip_management(in_raster=datFileName, rectangle=roughRectangle, out_raster=clipped, in_template_dataset="", nodata_value="", clipping_geometry="NONE", maintain_clipping_extent="NO_MAINTAIN_EXTENT")
            #print "projecting to state plane"
            arcpy.ProjectRaster_management(in_raster=clipped, out_raster=proj, out_coor_system="PROJCS['USA_Contiguous_Albers_Equal_Area_Conic_USGS_version',GEOGCS['GCS_North_American_1983',DATUM['D_North_American_1983',SPHEROID['GRS_1980',6378137.0,298.257222101]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]],PROJECTION['Albers'],PARAMETER['False_Easting',0.0],PARAMETER['False_Northing',0.0],PARAMETER['Central_Meridian',-96.0],PARAMETER['Standard_Parallel_1',29.5],PARAMETER['Standard_Parallel_2',45.5],PARAMETER['Latitude_Of_Origin',23.0],UNIT['Meter',1.0]]", resampling_type="NEAREST", cell_size="923.166496502035 923.166496502035", geographic_transform="NAD_1983_To_WGS_1984_1", Registration_Point="", in_coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]", vertical="NO_VERTICAL")
            
            #clip it to the extent of interest
            #print "clipping dataset"
            arcpy.gp.ExtractByMask_sa(proj, regionClipShape, regionclip)
            
            print "setting nulls"
            arcpy.gp.SetNull_sa(regionclip, regionclip, swe, '"Value" <0')
            
            #clean up!
            os.remove(datFileName+".gz")
            arcpy.Delete_management(datFileName)
            arcpy.Delete_management(clipped)
            arcpy.Delete_management(proj)
            arcpy.Delete_management(regionclip)
        else:
            print "Skipping SWE file for "+ datestamp +", already done"
            
            
# Calculates statistics by month. Note that in the file naming conventions, months are 1 or 2 digits, so Jan is 1 while Nov is 11:
# Number of days in that month where snow was <8inches - mds_<YYYY><M>.tif
# Number of days in that month where there was no snow - mdns_<YYYY><M>.tif
# Average depth of snow during the month - depth_ave_<YYYY><M>.tif
# Standard deviation of the snow depth during the month - depth_std_<YYYY><M>.tif
# These files are output into the individual year and month folders of the processed SNODAS data
            
def monthlyDepth(theYear,theMonthNumb):
    arcpy.env.overwriteOutput=True
    theYear=str(theYear)
    theMonth=str(monthDict[theMonthNumb])
    
    print "going through depths for  "+theYear+" year and " + theMonth + " month"
    
    #build the source path
    theWorkingPath=workingDirectory+ theYear + "/" + theMonth + "/"
    if (os.path.isdir(theWorkingPath)!=True):
        print "No directory found for this year and month"
        return
    #Delete the previous verision if it already exists
    if(os.path.isfile("month_sum_"+theYear+str(theMonthNumb)+".tif")==True):
        arcpy.Delete_management("month_sum_"+theYear+str(theMonthNumb)+".tif")
        
    os.chdir(theWorkingPath)
    
    # get the list of tar files in this directory
    arcpy.env.workspace = theWorkingPath
    tiffs = arcpy.ListFiles('*.tif')
    cons=[]
    nosnowcons=[]
    stats=[]
    for tiff in tiffs:
        #Look only at files that match the naming pattern, as others could be in the mix.
        if (tiff[:8]=="vt_depth"):
            #add this file to the list of conditionals
            #The test is whether the raster was at or above 8 inches depth on this day, 1 if yes, 0 if no. SNODAS is in millimeters depth, so .2032m=1in
            cons.append("Con('"+tiff+"'>=203,1,0)")
            #Add to list of conditionals to count up days of no snow in the month, 1 if yes, 0 if no
            nosnowcons.append("Con('"+tiff+"'==0,1,0)")
            stats.append(tiff)
        
    #create conditional statements
    cond='+'.join(cons)
    nosnowcond='+'.join(nosnowcons)
    
    #run raster calculator to get monthly raster of days with snow>8inches (mds_yyyymm.tif) and hte number of days with no snow (mdns_yyyymm.tif)
    print "calculating number of days >8"
    #UNCOMMENT NEXT LINE TO RUN FROM SCRATCH
    #arcpy.gp.RasterCalculator_sa(cond, "mds_"+theYear+str(theMonthNumb)+".tif")
    print "calculating number of days with no snow"
    arcpy.gp.RasterCalculator_sa(nosnowcond, "mdns_"+theYear+str(theMonthNumb)+".tif")
    
    #calculate the mean and STD by month of snow depth using hte list of tifs
    print "calculating standard deviation of depth for the month"
    outCellStatistics = CellStatistics(stats, "STD", "DATA")
    outCellStatistics.save("depth_std_"+theYear+str(theMonthNumb)+".tif")
    print "calculating mean of the depth"
    outCellStatistics = CellStatistics(stats, "MEAN", "DATA")
    outCellStatistics.save("depth_ave_"+theYear+str(theMonthNumb)+".tif")
    
    print str(len(cons))+" days with data in month "+theMonth+" and year "+theYear
    

#Run through year/month combinations and extract files and/or process monthly summaries
for processYear in years:
    for processMonth in months:
        snodasExtractClip(directory,workingDirectory,processYear,processMonth)  
        #monthlyDepth(processYear,processMonth)
