# This code takes daily rasters of snow depth as derived from SNODAS in snodasProcessing.py and then transforms them into various water yearly summary rasters (meaning Nov/Dec of the year before and Jan-Apr of the year as named (eg a "Year" of 2010 is Nov/Dec 2009 and Jan-Apr 2010):
# Mean depth over all days of water year wy_mean_depth_<YYYY>.tif
# Standard deviation of snow depth over all days of water year inches wy_std_depth_<YYYY>.tif
# Sum of days over water year where depth >8 inches wy_sum_<YYYY>.tif
# Sum of days over water year where there was no snow wy_sum_nosnow_<YYYY>.tif
# Author: Jim Duncan, UVM/FEMC james.duncan@uvm.edu
# Last updated: 4/12/2019

import arcpy
from arcpy.sa import *
from arcpy import env
arcpy.CheckOutExtension("Spatial")
import os
import datetime
#********************for testing********************
import logging
logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')
#**********************************************
#List of water years to compute
waterYears=["2005","2006","2007","2008","2009","2010","2011","2012","2013","2014","2015","2016","2017","2018","2019","2020"]
#waterYears=["2006"]
#crosswalks month values to naming conventions in the filesystem
monthDict={"11":"11_Nov","12":"12_Dec","1":"01_Jan","2":"02_Feb","3":"03_Mar","4":"04_Apr","5":"05_May","6":"06_Jun","7":"07_Jul","8":"08_Aug","9":"09_Sep","10":"10_Oct"}
outPath="D:/Duncan_SNODAS/snodasProcessing/waterYearStats/"
sourceDir="D:/Duncan_SNODAS/snodasProcessing/"
durTemplate="D:/Duncan_SNODAS/durTemplate2.tif"

##End of user input##

durThisStepName=sourceDir+"durThisStep.tif"

#Set environment
arcpy.env.workspace=sourceDir
arcpy.env.overwriteOutput=True
os.chdir(sourceDir)

#Loop through each water year (oct of previous year to April of current year)
for waterYear in waterYears:
    print ("Working on water year "+waterYear)
    
    #Define arrays to hold lists of rasters
    cons=[]
    calcList=[] #Hols rasters of days of snow >1 inch for annual summary
    sweCalcList=[] #Holds rasters of SWE for annual summary
    depthCalcList=[] #Holds rasters of daily data that should be used for mean and STD calculation for annual summaries
    durArray=[] # Holds daily depth rasters in date order for calculating snowpack duration
    
    #Define names of final files per water year
    durFinalName=outPath+"wy_duration_"+waterYear+".tif" #Snow duration per year
    snowDaysFinalName=outPath+"wy_sum_snowdays_"+waterYear+".tif"
    meanDepthFinalName=outPath+"wy_mean_depth_"+waterYear+".tif"
    stdDepthFinalName=outPath+"wy_std_depth_"+waterYear+".tif"
    meanSWEFinalName=outPath+"wy_mean_swe"+waterYear+".tif"
    maxDepthFinalName=outPath+"wy_max_depth_"+waterYear+".tif"
    arcpy.CopyRaster_management(durTemplate,durFinalName)
    arcpy.CopyRaster_management(durTemplate,durThisStepName)
    
    print("Looping through months")
    for key, month in monthDict.items():
        #If the month is November or December, we tack this on to the previous water year, otherwise goes to water year matching the year value
        if int(key)>=11:
            newWaterYear=str(int(waterYear)-1)
        elif int(key)<=4:
            newWaterYear=waterYear
        #Only do stuff if we are between November and April
        if (int(key)<=4 or int(key)>=11):
            #Gather list of file names for calculating yearly mean depth and std
            arcpy.env.workspace=sourceDir+newWaterYear+"/"+monthDict[key]
            os.chdir(sourceDir+newWaterYear+"/"+monthDict[key])
            tiffs = arcpy.ListFiles('*.tif')
            for tiff in tiffs:
                if (tiff[:12]=="region_depth"):
                    #add this file to the list of rasters to compute depth from
                    depthCalcList.append(newWaterYear+"/"+monthDict[key]+"/"+tiff)
                    
                    #add this file to the list of conditionals for whether snow was over 1 inch that day
                    #The test is whether the raster was at or above 1 inch depth on this day, 1 if yes, 0 if no. SNODAS is in millimeters depth, so .0254m=1in
                    cons.append("Con('"+newWaterYear+"/"+monthDict[key]+"/"+tiff+"'>25,1,0)")
                    
                    #compare this day's depth to the tracking raster. If 
                if (tiff[:10]=="region_swe"):
                    #add this file to the list of rasters to compute SWE stats from
                    sweCalcList.append(newWaterYear+"/"+monthDict[key]+"/"+tiff)
            #loop through days in a month to build list in order of depth rasters for calculating snow duration
            print("Looping through days")
            for i in range(1,32): #range( )does not include the stop number, so have to go to 32
                 #Check that current day is in the dataset
                 if(os.path.isfile("region_depth_"+newWaterYear+key.zfill(2)+str(i).zfill(2)+".tif")==True) :
                    datestamp=newWaterYear+key.zfill(2)+str(i).zfill(2)
                    depthRaster=Raster("region_depth_"+datestamp+".tif")
                    durArray.append(depthRaster)
    print("Calculating stats for "+newWaterYear)
    #Switch back to main dir
    arcpy.env.workspace=sourceDir
    os.chdir(sourceDir)
    
    
    #make initial zero copies of duration template for tallying duration
    
    #Create raster objects for iterating
    durFinal=Raster(durFinalName)
    durTest=Raster(durTemplate)
    print(durTest.properties)
    durThisStep=Raster(durThisStepName)
    print(durFinal)
    #Set to readonly bein gfalse so we can edit them
    durFinal.readOnly=False
    durThisStep.readOnly=False
    
    for dayDepth in durArray:
        print (dayDepth.catalogPath)
        for i, j in dayDepth:
            #print (dayDepth[i,j])
            if dayDepth[i,j]>0:
                #print("We've got one!")
                durThisStep[i,j] += 1
            else:
                durThisStep[i,j]=0
            if durFinal[i,j]<durThisStep[i,j]:
                durFinal[i,j]=durThisStep[i,j]
    #Save the final duration raster
    durFinal.save()
    #Delete the raster objects and the durThisStep raster so it can be created again.
    del durFinal
    del durThisStep
    arcpy.Delete_management(durThisStepName)
    
    
    #create conditional statement for snow > 1 inch
    cond='+'.join(cons)
    #run raster calculator to get year raster of days with snow>1inch
    print ("calculating number of days >1")
    arcpy.gp.RasterCalculator_sa(cond, snowDaysFinalName)
    
    #Calculate cell statistics for mean SWE, mean depth, std depth, max depth. 
    print ("calculating mean of snow depth over  water year")
    outCellStatistics = CellStatistics(depthCalcList, "MEAN", "DATA")
    outCellStatistics.save(meanDepthFinalName)
    print ("calculating STD of snow depth over  water year")
    outCellStatistics = CellStatistics(depthCalcList, "STD", "DATA")
    outCellStatistics.save(stdDepthFinalName)
    print ("calculating max of snow depth over  water year")
    outCellStatistics = CellStatistics(depthCalcList, "MAXIMUM", "DATA")
    outCellStatistics.save(maxDepthFinalName)
    print ("calculating mean of SWE over  water year")
    outCellStatistics = CellStatistics(sweCalcList, "MEAN", "DATA")
    outCellStatistics.save(meanSWEFinalName)


# Replace a layer/table view name with a path to a dataset (which can be a layer file) or create the layer/table view within the script
# The following inputs are layers or table views: "snodas_ranPTs_GCMa2_19812010_MonthlyVals"
#arcpy.env.workspace="I:/Administration/jim_docs/McSten/Snowmobile/snodasProcessing/"
#arcpy.gp.Sample_sa(filelist, "I:/Administration/jim_docs/McSten/Snowmobile/snodasStatistics/snodas_ranPTs_GCMa2_19812010_MonthlyVals.shp", "I:/Administration/jim_docs/McSten/Snowmobile/snodasStatistics/MonthlySampleVals", "NEAREST", "FID", "CURRENT_SLICE")
